------------------------------------------------------------------
-- Comm3 ABI TO Comm2 utility, v1.0 BETA							--
-- by GrafZeppelin (ferdinand.graf.zeppelin@gmail.com)			--
------------------------------------------------------------------

utility abiImport "Comm3 ABI TO Comm2"
(
	
	--------------------
	-- User Interface --
	--------------------
	
	group "About"
	(
		label lab11 "Comm3 ABI TO Comm2"
		
		HyperLink addy "by GrafZeppelin" align:#center address:"mailto:ferdinand.graf.zeppelin@gmail.com" color:(color 0 0 170) hoverColor:(color 170 0 0)
	)
	
	group "Import"
	(
		checkBox extractTexturesCheckBox "Extract Textures" checked:true align:#left
		checkBox createFolderCheckBox "Create Textures Folder" checked:true align:#left enabled:true
		button importButton "Import..." width:80 height:25
	)
	
	----------------------
	-- Helper Functions --
	----------------------
	
	function write file value =
	(
		format "%" (value as string) to:file
	)
	
	function writeln file value =
	(
		format "%\n" (value as string) to:file
	)
	
	-- Read a C-String of a certain length
	function ReadCString stream length = 
	(
		local str = ""
		local b
		for i = 1 to length do
		(
		    b = ReadByte stream
			if b == 0 then
			(
				fseek stream (length - i) #seek_cur
				return str
			)
			str += bit.intAsChar b
		)
		return str
	)
	
	--------------------------
	-- Main Import Function --
	--------------------------
	
	on importButton pressed do
	(
		
		-- Open up a dialog box
		abiFileName = getOpenFileName caption:"Import ABI" \
					types:"Commandos 3 ABI (*.abi)|*.abi|All Files (*.*)|*.*|"
		
		-- If user made a selection, begin importing
		if abiFileName != undefined then
		(
			local currentPath = getFilenamePath abiFileName
			
			-- Open the file
			abiStream = fopen abiFileName "rb"
			
			local string_LDMB = 1112360012
			local string_1011 = 825307185
			
			local isValid = true
			
			/*
			try
			(
			*/
				-- Check identifier
				if (ReadLong abiStream) != string_LDMB then
				(
					messageBox "Not a valid ABI file!"
					isValid = false
				)
				-- Check version
				else if (ReadLong abiStream) != string_1011 then
				(
					messageBox "Not a Commandos 3 ABI file.\n\nTry to open it with the Commandos 2 importer!"
					isValid = false
				)
				else
				(
					
					local directory = ""
					if ( extractTexturesCheckBox.checked and createFolderCheckBox.checked ) then
					(
						directory = ( ( filenameFromPath abiFileName ) + ".files\\" )
						
					)
					
					local texName = (abiFileName+"_322.abi")
					texStream = fopen texName "wb"
					
					local num_000 = ReadShort abiStream #unsigned
					local num_timeaxis = ReadShort abiStream #unsigned
					local num_animation = ReadShort abiStream #unsigned
					local num_texture = ReadShort abiStream #unsigned
					local num_model = ReadShort abiStream #unsigned
					local num_bone = ReadShort abiStream #unsigned
					
					local setoff_texture = ReadLong abiStream #unsigned
					local setoff_model = ReadLong abiStream #unsigned
					local setoff_animation = ReadLong abiStream #unsigned
					local long_bone = ReadLong abiStream #unsigned
					local long_animation = ReadLong abiStream #unsigned
					

												
					writeLong texStream   1112360012 --LDMB
					writeLong texStream   808792113 --1050
					
					writeLong texStream num_timeaxis 
					writeLong texStream num_animation  
					writeLong texStream num_texture   
					
					


					
				
					for i = 1 to num_texture do
					(
					   local texture_width = #()
					   local texture_height = #()
						
						
						local texture_width = ReadShort abiStream #unsigned
						local texture_height = ReadShort abiStream #unsigned
						
						
						WriteLong texStream 0
						WriteLong texStream texture_width #unsigned
						WriteLong texStream texture_height #unsigned
						
						
						fseek abiStream 768 #seek_cur -- skip Color Palette						
							
							for j = 1 to 32 do
							(
								local texture_name = ReadByte abiStream
																
								WriteByte texStream texture_name
								
         				)

						fseek abiStream -800 #seek_cur -- go back to Color Palette
						
						
							for j = 1 to 256 do
							(
								local red = ReadByte abiStream
								local green = ReadByte abiStream
								local blue = ReadByte abiStream
								
								WriteByte texStream red
								WriteByte texStream green
								WriteByte texStream blue
								
							)

						fseek abiStream 32 #seek_cur
						
						
						local texture_size= ReadLong abiStream #unsigned
						local texture_offset = ReadLong abiStream #unsigned 
						
						
						
						fseek abiStream texture_offset #seek_set -- go  to Color 
															
						

						for j = 1 to texture_height*texture_width do
							(
								local bitmap = ReadByte abiStream
																
								WriteByte texStream bitmap
							)
							
                fseek abiStream (40+i*812) #seek_set -- go  to  
							

											
						)
						
						

					fseek abiStream (num_bone*52) #seek_cur	
						
			
					
		
					
					writeLong texStream num_model 
					writeLong texStream num_bone 	
					
					
					local model_num_vertex = #()
					local model_num_polygon = #()
					local model_offset_texure_m = #()
					local model_name = #()
					local model_offset_model_m = #()
					
					
					
					for i = 1 to num_model do
					(
						local model_num_vertex = ReadShort abiStream #unsigned 
						local model_num_polygon = ReadShort abiStream #unsigned 
						
			   		writeLong texStream model_num_vertex 
	    			writeLong texStream model_num_polygon 	
						
						
						local model_offset_texure_m = ReadShort abiStream #unsigned 
						
						
							
							for j = 1 to 32 do
							(
								local model_name  = ReadByte abiStream
																
								WriteByte texStream model_name 
								
         				)

						
						local model_offset_model_m = ReadLong abiStream #unsigned 
						
						fseek abiStream model_offset_model_m #seek_set
						
							for j = 1 to model_num_vertex do
						(
							local x = ReadFloat abiStream
							local y = ReadFloat abiStream
							local z = ReadFloat abiStream
							
							WriteFloat texStream x
							WriteFloat texStream y
							WriteFloat texStream z
							
						)
						
							for j = 1 to model_num_polygon do
						(
							

							
							local num_borders = ( ReadByte abiStream #unsigned )						
							local texture_id = ( ( ReadShort abiStream #unsigned )  )
							
							
							WriteByte texStream num_borders 
							WriteByte texStream texture_id
							 
							for k = 1 to num_borders*6 do
							(
								
								local model_poly  = ReadByte abiStream
																
								WriteByte texStream model_poly
				
							)
					
							
						)
						
						
						for j = 1 to num_bone do
						(
						
							local StartVertex  = ReadLong abiStream #unsigned
							local EndVertex  = ReadLong abiStream #unsigned
							
							WriteLong texStream StartVertex #unsigned
					  	WriteLong texStream EndVertex #unsigned
						)
						
						
						 fseek abiStream (40+num_texture*812+num_bone*52+42*i) #seek_set
						
						
					)
					
					fseek abiStream (-num_bone*52-num_model*42 ) #seek_cur	
					
					
					for i = 1 to num_bone do
					(
						local boneParent = ReadLong abiStream #unsigned
						
						writeLong texStream boneParent #unsigned
						
						local bone00= ReadLong abiStream #unsigned
						
												
							for j = 1 to 44 do
							(
								local bone_xyz_name = ReadByte abiStream
																
								WriteByte texStream bone_xyz_name
								
         			)
         			
         		writeLong texStream bone00 #unsigned

					)
						 
					
					fseek abiStream (num_model*42 +num_animation*52+long_bone) #seek_cur	
					
					


					
					for i = 1 to num_timeaxis do
					   (
						local num_translateKeyframes = ReadByte abiStream #unsigned
						local num_rotateKeyframes = ReadByte abiStream #unsigned
						
				   	
							local dx =  #()
							local dy =  #()
							local dz =  #()
							local rx =  #()
							local ry =  #()
							local rz =  #()	
						
								
						
						if num_translateKeyframes != 0 then
					
						(
						
						WriteLong texStream num_translateKeyframes #unsigned
						
						for j= 1 to num_translateKeyframes do
						(
							
							append dx ( readShort abiStream #signed)
							append dy ( readShort abiStream #signed)
							append dz ( readShort abiStream #signed)
							
						)
						
						local t_timestamp =  #()							
						local r_timestamp =  #()
						
							
							
						for j = 1 to num_translateKeyframes do
						(
							append  t_timestamp ( ReadByte abiStream #unsigned)
							
							WriteByte texStream t_timestamp[j] #unsigned
							
							WriteShort texStream dx[j] #signed
							WriteShort texStream dy[j] #signed
							WriteShort texStream dz[j] #signed
							
						)								
						
						
						if num_rotateKeyframes != 0 then
						
						(
					
						Writelong texStream num_rotateKeyframes #unsigned
	
							
						for k = 1 to num_rotateKeyframes do
						(
							
								append rx (( ReadByte abiStream #signed))
								append ry (( ReadByte abiStream #signed))
								append rz (( ReadByte abiStream #signed))
						
								--local euler =   eulerAngles  rx  ry   rz  
													
						)
						
						
						
						for k= 1 to num_rotateKeyframes do
							(
								append  r_timestamp (ReadByte abiStream #unsigned)
								
						   	WriteByte texStream r_timestamp[k] #unsigned
							
								local rw =sqrt(128*128- rx[k]*rx[k]-ry[k]*ry[k]-rz[k]*rz[k])
									
								writeShort texStream ( rx[k] * 256)
								writeShort texStream ( ry[k] * 256)
								writeShort texStream ( rz[k] * 256)
								writeShort texStream ( rw *	(256))
								
									)
							
								
						)
						
						else 
						(
						
										            
	            Writelong texStream 2 #unsigned	            
	            
	            WriteByte texStream 0 #unsigned
	            
	            Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 32767 #signed
							
							local rzg =  t_timestamp[num_translateKeyframes]
							WriteByte texStream rzg  #unsigned
							
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 32767 #signed
						
						)
						
					)
				else 	
					(
				   	if num_rotateKeyframes != 0 then
					
			     (
						
						
						local r_timestamp =  #()
						for k = 1 to num_rotateKeyframes do
						(
							
								append rx (( ReadByte abiStream #signed))
								append ry (( ReadByte abiStream #signed))
								append rz (( ReadByte abiStream #signed))
									
						)
						
							
						for j = 1 to num_rotateKeyframes do
							(
									append  r_timestamp( ReadByte abiStream #unsigned)
								
							)
					
						(
						            
	            Writelong texStream 2 #unsigned	            
	            
	            WriteByte texStream 0 #unsigned
	            
	            Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							
							local rzg3 = r_timestamp[num_rotateKeyframes]
							WriteByte texStream rzg3 #unsigned
							
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							
						)
						
						
						Writelong texStream num_rotateKeyframes #unsigned
						
						
						for k = 1 to num_rotateKeyframes do
							(
									
								WriteByte texStream r_timestamp[k] #unsigned
									
								local euler = eulerAngles rx[k] ry[k] rz[k]
								
								local q = eulerToQuat euler order:1
								
								local rw =sqrt(128*128- rx[k]*rx[k]-ry[k]*ry[k]-rz[k]*rz[k])
									
								writeShort texStream ( rx[k] * 256)
								writeShort texStream ( ry[k] * 256)
								writeShort texStream ( rz[k] * 256)
								writeShort texStream ( rw *	(256))
									
							)
					
					)
					else 	
					(
					
										            
	            Writelong texStream 2 #unsigned	            
	            
	            WriteByte texStream 0 #unsigned
	            
	            Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							
							
							WriteByte texStream 1 #unsigned
							
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
					
						  Writelong texStream 2 #unsigned	            
	            
	            WriteByte texStream 0 #unsigned
	            
	            Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 32767 #signed
							
							
							WriteByte texStream 1 #unsigned
							
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 0 #signed
							Writeshort texStream 32767 #signed
					
				  	)
				
					)
						
					
						
						if mod num_translateKeyframes  2 != 0  then
						--if ( bit.and num_translateKeyframes 1 ) != 0 then
						(
							local timeaxis64 = ReadByte abiStream #unsigned
						)
						
					)
					
					
			    fseek abiStream setoff_animation #seek_set		
					
					
					local animation_namel=  #()		
					local animation_nameul=  #()
					
					for i = 1 to num_animation do
					(
					     local animation_name =  #()
							
							 for j = 1 to 36 do
							  (
							    	append  animation_name ( ReadByte abiStream)
										
         		   	)
         		  
						   	append animation_namel animation_name
									
								
								local animation_nameu =    #()
							
								for j = 1 to 16 do
				  			(
								append animation_nameu  ( ReadByte abiStream)
														
							
           			)
           			append animation_nameul animation_nameu
         			
						
					)
				
					for i = 1 to num_animation do
					
					(    
						
					  	for j = 1 to 36 do
								(
									WriteByte texStream ( animation_namel[i][j] )
								)
						
						
				 			writeLong texStream -1
         			writeLong texStream -1
         			
         			 for j = 1 to 16 do
								(
									WriteByte texStream ( animation_nameul[i][j] )
								)
						
				 			
				 			
				 			local num_related_bones = ReadLong abiStream #unsigned
				 							 			
				 			writeLong texStream num_related_bones
         				
						for i = 1 to num_related_bones*8 do
						(
								local num_related_boness  = ReadByte abiStream
																
								WriteByte texStream num_related_boness
						)
						
						
					)
					
				
					
					fclose texStream
				)
			
			/*
			)
			catch
			(
				messageBox "Error reading ABI file!"
				isValid = false
				fclose abiStream
				throw
			)
			*/
			
			-- Close the file stream
			fclose abiStream
			
			if isValid then
			(
				-- Redraw max viewports
				max views redraw
				
				print "Import completed."
			)
		)
	)
	
	on extractTexturesCheckBox changed nowChecked do
	(
		createFolderCheckBox.enabled = nowChecked
	)
)
