------------------------------------------------------------------
-- Commandos 2 ABI Import utility, v1.0 BETA (18/8/11)			--
-- by GrafZeppelin (ferdinand.graf.zeppelin@gmail.com)			--
-- 2013 2 11 ͯ5099fixed                                               --
------------------------------------------------------------------

rollout abiImport "Commandos 2 ABI Import" width:184 height:299
(
	
	--------------------
	-- User Interface --
	--------------------
	
	GroupBox grp1 "About" pos:[4,7] width:176 height:76
	label lab11 "Commandos 2" pos:[59,26] width:67 height:13
	label lab12 "ABI Importer v1.2013 BETA" pos:[30,44] width:120 height:13
	HyperLink addy "by GrafZeppelin" pos:[55,62] width:75 height:15 address:"mailto:ferdinand.graf.zeppelin@gmail.com" color:(color 0 0 170) hovercolor:(color 170 0 0)
		

	GroupBox grp2 "Import" pos:[4,88] width:176 height:158
	checkbox extractTexturesCheckBox "Extract Textures" pos:[13,107] width:101 height:15 checked:true
	checkbox createFolderCheckBox "Create Textures Folder" pos:[13,127] width:131 height:15 enabled:true checked:true
	button importButton "Import..." pos:[52,147] width:80 height:25
	
	spinner modelSpinner "Model: " pos:[20,184] width:137 height:16 range:[1,100,1] type:#integer scale:1
	
	spinner animationSpinner "Animation: " pos:[23,217] width:136 height:16 range:[1,100,2] type:#integer scale:1
	----------------------
	-- Helper Functions --
	----------------------
	
	-- Read an unsigned byte
	
		function write file value =
		(
			format "%" (value as string) to:file
		)
		
		function writeln file value =
		(
			format "%\n" (value as string) to:file
		)

	
	function ReadUByte stream =
	(
		local byte = ReadByte stream
		if byte < 0 then
		(
			byte += 256
		)
		return byte
	)
	
	-- Read an unsigned integer little-endian
	function ReadUInt stream =
	(
		local int = (ReadUByte stream) + (ReadUByte stream) * 256 + (ReadUByte stream) * 65536 + (ReadUByte stream) * 16777216
		return int
	)
	
	-- Read an unsigned short little-endian
	function ReadUShort stream =
	(
		local short = (ReadUByte stream) + (ReadUByte stream) * 256
		return short
	)
	
	-- Read an signed short little-endian
	function ReadSShort stream =
	(
		local short = ReadUShort stream
		if short > 32767 then short -= 65536
		return short
	)
	
	-- Read a C-String of a certain length
	function ReadCString stream length = 
	(
		local str = ""
		local b
		for i = 1 to length do
		(
		b = ReadByte stream
			if b == 0 then
			(
				fseek stream (length - i) #seek_cur
				return str
			)
			str += bit.intAsChar b
		)
		return str
	)
		fn setNodeWorldRotation theNode theRot = 
	(
		in coordsys (transmatrix theNode.transform.pos)
			theNode.rotation = theRot
	)
	
	--------------------------
	-- Main Import Function --
	--------------------------
	
	
	on extractTexturesCheckBox changed nowChecked do
	(
		createFolderCheckBox.enabled = nowChecked
	)
	on importButton pressed do
	(
		
		-- Open up a dialog box
		abiFileName = getOpenFileName caption:"Import ABI" \
					types:"Commandos 2 ABI (*.abi)|*.abi|All Files (*.*)|*.*|"
		
		-- If user made a selection, begin importing
		if abiFileName != undefined then
		(
			local currentPath = getFilenamePath abiFileName
			
			-- Open the file
			abiStream = fopen abiFileName "rb"
			
			local string_LDMB = 1112360012
			local string_1040 = 808726577
			local string_1060 = 808857649
			
			local isValid = true
			
			clearlistener()
			
			
			
			try
			(
				-- Check identifier
				if (ReadLong abiStream) != string_LDMB then
				(
					messageBox "Not a valid ABI file!"
					isValid = false
				)
				else 
				(
					local version = ReadLong abiStream
					local num_timeaxis = ReadUInt abiStream
					local num_animation = ReadUInt abiStream
					local num_texture = ReadUInt abiStream
					
					local newMaterial = multimaterial numsubs:num_texture
					meditmaterials[1] = newMaterial
					
					local directory = ""
					if ( extractTexturesCheckBox.checked and createFolderCheckBox.checked ) then
					(
						directory = ( ( filenameFromPath abiFileName ) + ".files\\" )
						makeDir ( currentPath + directory )
					)
							
		
		
					local StartVertex  = #()
					local EndVertex   = #()
			
					local show_animation = animationSpinner.value
					
					for i = 1 to num_texture do
					(
						local block_identifier = ReadUInt abiStream
						local width = ReadUInt abiStream
						local height = ReadUInt abiStream
						local name = ReadCString abiStream 32
						
					
						-- Assign Bitmap
						newMaterial[i] = standard name:name
						newMaterial[i].diffuseMap = BitmapTexture()
						newMaterial[i].diffuseMap.name = name
						newMaterial[i].diffuseMap.filename = ( currentPath + directory + name )
						
						-- Show Map in Viewport
						showTextureMap newMaterial[i] true
						
						-- Extract Bitmap
						if extractTexturesCheckBox.checked then
						(
							local texName = currentPath + directory + name
							texStream = fopen texName "wb"
							
							-- Bitmap Header
							writeByte texStream 0x42
							writeByte texStream 0x4D
							writeLong texStream ( 1078 + width * height )
							writeLong texStream 0
							writeLong texStream 1078
							writeLong texStream 40
							writeLong texStream width
							writeLong texStream height
							writeShort texStream 1
							writeShort texStream 8
							writeLong texStream 0
							writeLong texStream ( width * height)
							writeLong texStream 0
							writeLong texStream 0
							writeLong texStream 0
							writeLong texStream 0
							
							-- Color Palette
							for j = 1 to 256 do
							(
								local red = readByte abiStream
								local green = readByte abiStream
								local blue = readByte abiStream
								
								writeByte texStream blue
								writeByte texStream green
								writeByte texStream red
								writeByte texStream 0
							)
							
							local bitmap = #()
							
							for j = 1 to height do
							(
								local line = #()
								for k = 1 to width do
								(
									append line ( readByte abiStream )
								)
								append bitmap line
							)
							
							for j = 1 to height do
							(
								for k = 1 to width do
								(
									writeByte texStream ( bitmap[height-j+1][k] )
								)
							)
							
							fclose texStream
							
						)
						else
						(
							-- Skip Bitmap
							fseek abiStream ( 768 + width * height ) #seek_cur
						)
						
					)
	      	
	    
					
					if version == string_1060 then
					(
						local identifier = ReadUByte abiStream
					)
							-- FILE OPERATIONS --
					f = createfile (abiFileName+"test.txt")
		
	   				writeln f "This is just a test!"
	
	
					local num_model = ReadUInt abiStream
					local num_bone = ReadUInt abiStream
					
					local vertss = #()
					local modelnames = #()	
					for i = 1 to num_model do
					(
						local verts = #()
						local faces = #()
						
						local texCoords = #()
						local texFaces = #()
						local borders_sum = 0
						
						local faceMaterialIDs = #()
						
						local num_vertex = ReadUInt abiStream
						local num_polygon = ReadUInt abiStream
						local name = ReadCString abiStream 32
						append modelnames name
						
	
						
						for j = 1 to num_vertex do
						(
							local x = ReadFloat abiStream
							local y = ReadFloat abiStream
							local z = ReadFloat abiStream
							append verts [ x, y, z ]
							
						)
						vertss =  verts
						
						if version == string_1040 then
						(
							for j = 1 to num_polygon do
							(
								local num_borders = ReadUInt abiStream
								local texture_id = ( ( ReadUInt abiStream) + 1 )
								
								local ID = #()
								local UV = #()
								
								for k = 1 to num_borders do
								(
									append ID ( ( ReadUInt abiStream ) + 1 )
									append UV ( borders_sum + k )
									local u = ( ReadFloat abiStream )
									local v = ( ( ReadFloat abiStream ) * -1 )
									append texcoords [ u, v, 1 ]
								)
								
								for k = 1 to num_borders-2 do
								(
									append faces [ ID[k+2], ID[k+1], ID[1] ]
									append texFaces [ UV[k+2], UV[k+1], UV[1] ]
									append faceMaterialIDs texture_id
								)
								
								borders_sum += num_borders
							)
						)
						else -- version != 1040
						(
							for j = 1 to num_polygon do
							(
								local num_borders = ReadUByte abiStream
								local texture_id = ( ( ReadUByte abiStream) + 1 )
								
								local ID = #()
								local UV = #()
								
								for k = 1 to num_borders do
								(
									append ID ( ( ReadUShort abiStream ) + 1 )
									append UV ( borders_sum + k )
									local u = ( ( ReadSShort abiStream ) / 4096.0 )
									local v = ( ( ReadSShort abiStream ) / -4096.0 )
									append texcoords [ u, v, 1 ]
								)
								
								for k = 1 to num_borders-2 do
								(
									append faces [ ID[k+2], ID[k+1], ID[1] ]
									append texFaces [ UV[k+2], UV[k+1], UV[1] ]
									append faceMaterialIDs texture_id
								)
								
								borders_sum += num_borders	
							)
						)
						
						-- Skip Bone Data
						-- fseek abiStream ( num_bone * 8 ) #seek_cur
					
	     
		
					for j = 1 to num_bone do
					(
	
						append StartVertex  ( ReadUInt abiStream )
						append EndVertex    ( ReadUInt abiStream )
	
					
					
						--writeShort texStream StartVertex[j] 
						--writeShort texStream EndVertex[j] 	
							
						write f StartVertex[j] 
						write f "	"
						writeln f EndVertex[j] 
					)
						
					writeln f "	"
					writeln f "	"	
					if i ==	modelSpinner.value	do
					(
	
					-- Construct the mesh object
					local newMesh = mesh name:name pos:[0,0,0] \
									vertices:verts faces:faces \
									material:newMaterial
					
					-- Set face material IDs
					for j = 1 to faceMaterialIDs.count do
					(
						setFaceMatID newMesh j faceMaterialIDs[j]
					)
					
					-- Add texcoords
					if texcoords.count > 0 do
					(
						-- Add texcoords
						setNumTVerts newMesh texcoords.count
						
						for j = 1 to texcoords.count do
						(
							setTVert newMesh j texcoords[j]
						)
						
						-- Build texcoord faces
						buildTVFaces newMesh
						
						-- Set texcoord faces
						for j = 1 to texFaces.count do
						(
							setTVFace newMesh j texFaces[j]
						)
						
					)
				
					-- Update mesh
					update newMesh
					)
					
				)
					
						
				local bone_n = #( )
					
				for i = 1 to num_bone do
				(
												
					local  Parent_Index =ReadUInt abiStream
					local x = ReadFloat abiStream
					local y = ReadFloat abiStream
					local z = ReadFloat abiStream
					local  bone_name =(( ReadCString abiStream 32) as String)
					local xxx = ReadUInt abiStream
				
							
					append bone_n ( BoneSys.createBone [ x, y, z ] [ x+10, y, z ][ x, y, z ])
					scale bone_n[i] [0.3,0.3,0.3]
	
					in coordsys parent 	bone_n[i].rotation=	(quat 0 0 0 1)
					bone_n[i].pos= [ x, y, z ] 
	
					bone_n[i].name = bone_name										
					
				
					if Parent_Index != -1 then bone_n[i].parent = bone_n[Parent_Index+1] 
					in coordsys parent 	bone_n[i].rotation=	(quat 0 0 0 1)
					
				
					
					
				/*
					if Parent_Index == -1 then
					(
					bone_n [i] =	bone pos:[ x, y, z ]
					)
					else
					
					(
						in bone_n[Parent_Index+1] bone_n[i] = bone pos:[ x, y, z ]
					)							
					
	        bone_n[i].name == bone_name
					append bone_n bone_n[i]	
	
					
					
					writeFloat texStream x
					writeFloat texStream y
					writeFloat texStream z
				*/
					write f Parent_Index
					write f "		"	
					write f x
					write f "		"						
					write f y
					write f "		"
					write f z
					write f "		"
					writeln f bone_name
	
		
				)
	      max modify mode 
	  		vSkin = skin() 
				
				m= getNodeByName modelnames[modelSpinner.value]
				select m         
				convertToPoly m  
	-- node = getNodeByName boneNames
	
				addModifier m vSkin 
	
				
				for i=1 to bone_n.count do skinOps.addbone vSkin (getNodeByName bone_n[i].name ) -1
		
				 skinOps.selectBone vskin 1
					--for i=1 to vertss.count do skinOps.ReplaceVertexWeights vskin i 1 1
				step = "Skinning complete"
	
	
				for i=1 to bone_n.count do
				(
					if StartVertex[i] != EndVertex[i] then				
					(
						for j = (StartVertex[i]+1) to (EndVertex[i]) do
						(
								skinOps.ReplaceVertexWeights vSkin j  i 1		
						)
					)
				)
			
			skinOps.updateMirror vSkin
			
	/*		update m
		step = "Skinning complete"
		bi = #() --bone index array
	  wv = #() --weight value array
	 	append bi bone_n[i]
	  append wv 1.0
	
		for j = 1 to w.boneids.count do
		(
		  boneid = w.boneids[j]
		  weight = w.weights[j]
		  
		)   
		
		skinOps.ReplaceVertexWeights skinMod i bi wv
		for i = 1 to bone_n.count do	
		(
				if StartVertex[i]  != EndVertex[i] then
				(
					print EndVertex[i]-StartVertex[i]  
				)
	   )
	
	  )
	
	  max modify mode 
	  select Mesh 
	  testSkin = skin() 
	  addModifier Mesh testSkin 	
	  skinOps.addBone testSkin bone_n  1 	  
	  skinOps.selectBone testSkin 1 				
		for i = 1 to num_bone do	
	  (
		if StartVertex[i]  != EndVertex[i] then
		(
			for j = StartVertex[i]+1 to EndVerte[i] do
			(
				skinOps.setVertexWeights testSkin j bone_n[i] 1.0 		
			)	
		)
	 )
	
	*/  
						local Keyframes    = #()
	
						for i = 1 to num_timeaxis do
						(
							local tKeyframes      = #()
							local rKeyframes      = #()
							
							local num_translateKeyframes = ReadUInt abiStream
								
							writeln f "	"
							writeln f num_translateKeyframes
													
							for j = 1 to num_translateKeyframes do
							(
							
								local t_times = readuByte abiStream
								
								local dx =( readSShort abiStream ) /256.0
								local dy =( readSShort abiStream ) /256.0
								local dz =( readSShort abiStream ) /256.0
								
								append tKeyframes #(t_times ,[dx ,dy,dz ] ) ---
								
								write f t_times
								write f "	"
								write f dx
								write f "	"
								write f dy
								write f "	"
								writeln f dz
						
							)
							
							
							local num_rotateKeyframes = ReadUInt abiStream
							
							writeln f "	"
							writeln f num_rotateKeyframes						
									
							for j = 1 to num_rotateKeyframes do
							(
							
								local r_times = readuByte abiStream
								
								local rx =( readsshort abiStream )/32767.0
								local ry =( readsshort abiStream )/32767.0
								local rz =( readsshort abiStream )/32767.0
								local rw =( readsshort abiStream )/32767.0
								
								append rKeyframes #(r_times ,(quat rx ry  rz rw ) )
							
								aaa=(Quat rx ry  rz  rw) as eulerAngles
	
								write f r_times
								write f "	"
								write f rx
								write f "	"
								write f ry
								write f "	"
								write f rz
								write f "	"
								writeln  f rw
								
								--write f "	"
								--writeln f aaa
								
							)
							append Keyframes #(tKeyframes ,rKeyframes )
					)
					print Keyframes.count
					local boneani    = #()
	
					print "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz"
					for i = 1 to num_animation do
					(
					
						local name_animation = ReadCString abiStream 32
						local animation_timelong =  ReadUInt abiStream 
			      local name_animation_unknow = ReadCString abiStream 24			            
			      local num_related_bones = ReadUInt abiStream 
	     			
	     			write f name_animation
						write f "	"
						writeln f animation_timelong
	     			
	
	     			local bone_anim    = #()
	     			for j = 1 to num_related_bones do
						(
										
							local bone_id = ReadUInt abiStream 
							local transform_time_axis_idx = ReadUInt abiStream 
	        				
							append  bone_anim #(bone_id  ,transform_time_axis_idx  )
	
							write f bone_id
							write f "	"
							write f transform_time_axis_idx
	        		write f "	"
	
						)
						writeln f "	"
						append   boneani  bone_anim
						
						if i == show_animation then
						(
							frameRate = 24
							animationRange = interval 0 animation_timelong
							animate on
							(
								for j = 1 to num_related_bones do
								(
									in coordsys parent op = copy bone_n[bone_anim[j][1]+1].pos
									in coordsys parent oaa = bone_n[bone_anim[j][1]+1].rotation --in coordsys  local 
									 obb= (quat 0 0 0 1)
	
									for k = 1 to Keyframes[bone_anim[j][2]+1][1].count do
									(
										at time Keyframes[bone_anim[j][2]+1][1][k][1]  in coordsys parent  bone_n[bone_anim[j][1]+1].pos=op+ Keyframes[bone_anim[j][2]+1][1][k][2]
									)
									for k = 1 to Keyframes[bone_anim[j][2]+1][2].count do
									(
										 at time Keyframes[bone_anim[j][2]+1][2][k][1] in coordsys local  bone_n[bone_anim[j][1]+1].rotation= obb * inverse Keyframes[bone_anim[j][2]+1][2][k][2] --(quat 0 0 0 1 ) 
										 obb=Keyframes[bone_anim[j][2]+1][2][k][2]
										--WRONG-- in coordsys parent  bone_n[bone_anim[j][1]+1].rotation = inverse Keyframes[bone_anim[j][2]+1][2][k][2]
										--OR-- at time Keyframes[bone_anim[j][2]+1][2][k][1]  bone_n[bone_anim[j][1]+1].controller.rotation = Keyframes[bone_anim[j][2]+1][2][k][2] --(quat 0 0 0 1 ) 
	
	
	
									)
								
								)
							)
						
						)
						
					)
			
				close f
						
				--fclose texStream
				)
			
			)
			catch
			(
				messageBox "Error reading ABI file!"
				isValid = false
				fclose abiStream
				throw;	;			
			)
			
			-- Close the file stream
			fclose abiStream
			
			if isValid then
			(
				-- Redraw max viewports
				max views redraw
				
				print "Import completed."
			)
			messageBox  "Import completed"
		)
	)
)

CreateDialog   abiImport  202 284 25 25

