------------------------------------------------------------------
-- Commandos 3 ABI Import utility, v1.1							--
-- by GrafZeppelin (ferdinand.graf.zeppelin@gmail.com)			--
------------------------------------------------------------------

utility abiImport "Comm3 ABI Import" width:200 height:309
(
	
	--------------------
	-- User Interface --
	--------------------
	
	GroupBox grp1 "About" pos:[4,7] width:188 height:76
	label lab11 "Commandos 3" pos:[48,26] width:67 height:13
	label lab12 "ABI Importer v1.1" pos:[38,44] width:87 height:13
	HyperLink addy "by GrafZeppelin" pos:[44,62] width:75 height:15 address:"mailto:ferdinand.graf.zeppelin@gmail.com" color:(color 0 0 170) hovercolor:(color 170 0 0)
	
	GroupBox grp2 "Import" pos:[4,88] width:187 height:201
	checkbox extractTexturesCheckBox "Extract Textures" pos:[13,107] width:101 height:15 checked:true
	checkbox convertCheckBox "Convert to C2 ABI" pos:[10,140] width:128 height:16
	
	spinner modelSpinner "Model:       " pos:[17,160] width:137 height:16 range:[1,100,1] type:#integer scale:1
	spinner animationSpinner "Animation: " pos:[18,191] width:136 height:16 range:[1,100,2] type:#integer scale:1
	
	
	button importButton "Import..." pos:[8,216] width:128 height:25
	
	----------------------
	-- Helper Functions --
	----------------------
	
	-- Write a String which is not terminated by 0
	function WriteString file string =
	(
		for i = 1 to string.count do
		(
			WriteByte file ( bit.charAsInt string[i] )
		)
	)
	
	-- Write a String of a certain length
	function WriteCString file string length =
	(
		if string.count > length then
		(
			for i = 1 to length do
				WriteByte file ( bit.charAsInt string[i] )
			
			return 0
		)
		
		for i = 1 to string.count do
			WriteByte file ( bit.charAsInt string[i] )
		
		for i = string.count + 1 to length do
			WriteByte file 0
		
	)
	
	-- Read a C-String of a certain length
	function ReadCString stream length = 
	(
		local str = ""
		local b
		for i = 1 to length do
		(
		    b = ReadByte stream
			if b == 0 then
			(
				fseek stream (length - i) #seek_cur
				return str
			)
			str += bit.intAsChar b
		)
		return str
	)
	fn setNodeWorldRotation theNode theRot = 
	(
		in coordsys (transmatrix theNode.transform.pos)
			theNode.rotation = theRot
	)
	--------------------------
	-- Main Import Function --
	--------------------------
	
	on importButton pressed do
	(
		
		-- Open up a dialog box
		abiFileName = getOpenFileName caption:"Import ABI" \
					types:"Commandos 3 ABI (*.abi)|*.abi|All Files (*.*)|*.*|"
		
		-- If user made a selection, begin importing
		if abiFileName != undefined then
		(
			local currentPath = getFilenamePath abiFileName
			
			-- Open the file
			abiStream = fopen abiFileName "rb"
			
			local string_LDMB = 1112360012
			local string_1011 = 825307185
			
			local isValid = true
			
			--/* DEBUG
			try
			(
			--*/
				-- Check identifier
				if (ReadLong abiStream) != string_LDMB then
				(
					messageBox "Not a valid ABI file!"
					isValid = false
				)
				-- Check version
				else if (ReadLong abiStream) != string_1011 then
				(
					messageBox "Not a Commandos 3 ABI file.\n\nTry to open it with the Commandos 2 importer!"
					isValid = false
				)
				else
				(
					
					local directory = ( ( filenameFromPath abiFileName ) + ".files\\" )
					
					if ( extractTexturesCheckBox.checked or convertCheckBox.checked ) then
						makeDir ( currentPath + directory )
					
					local outStream
					if convertCheckBox.checked then
						outStream = fopen ( currentPath + directory + ( filenameFromPath abiFileName ) ) "wb"
					
					-- GENERAL INFO --
					local num_000 = ReadShort abiStream #unsigned
					local num_timeaxis = ReadShort abiStream #unsigned
					local num_animation = ReadShort abiStream #unsigned
					local num_texture = ReadShort abiStream #unsigned
					local num_model = ReadShort abiStream #unsigned
					local num_bone = ReadShort abiStream #unsigned
					
					local offset_texture = ReadLong abiStream #unsigned
					local offset_model = ReadLong abiStream #unsigned
					local offset_animation = ReadLong abiStream #unsigned
					local length_bone = ReadLong abiStream #unsigned
					local length_timeaxis = ReadLong abiStream #unsigned
					
					if convertCheckBox.checked then
					(
						writeString outStream "LDMB"
						writeString outStream "1050"
						writeLong outStream num_timeaxis
						writeLong outStream num_animation
						writeLong outStream num_texture
					)
					
					-- TEXTURES HEADER --
					local newMaterial = multimaterial numsubs:num_texture
					if extractTexturesCheckBox.checked then meditmaterials[1] = newMaterial
					
					local texture_width = #()
					local texture_height = #()
					local texture_name = #()
					local texture_size = #()
					local texture_offset = #()
					
					for i = 1 to num_texture do
					(
						
						append texture_width ( ReadShort abiStream #unsigned )
						append texture_height ( ReadShort abiStream #unsigned )
						
						fseek abiStream 768 #seek_cur -- skip Color Palette
						append texture_name ( ReadCString abiStream 32 )
						append texture_size ( ReadLong abiStream #unsigned )
						append texture_offset ( ReadLong abiStream #unsigned )
						fseek abiStream -808 #seek_cur -- go back to Color Palette
						
						local name = texture_name[i]
						
						if convertCheckBox.checked then
						(
							writeLong outStream 0
							writeLong outStream texture_width[i]
							writeLong outStream texture_height[i]
							writeCString outStream name 32
							
							-- get current pointer position
							local pos = ftell abiStream
							
							for j = 1 to 768 do
							(
								WriteByte outStream ( ReadByte abiStream )
							)
							
							-- go to the texture data
							fseek abiStream texture_offset[i] #seek_set
							
							for j = 1 to ( texture_width[i] * texture_height[i] ) do
							(
								WriteByte outStream ( ReadByte abiStream )
							)
							
							-- go back
							fseek abiStream pos #seek_set
						)
						
						if extractTexturesCheckBox.checked then
						(
							local width = texture_width[i]
							local height = texture_height[i]
							local texName = currentPath + directory + name
							texStream = fopen texName "wb"
							
							-- Bitmap Header
							WriteByte texStream 0x42
							WriteByte texStream 0x4D
							WriteLong texStream ( 1078 + width * height )
							WriteLong texStream 0
							WriteLong texStream 1078
							WriteLong texStream 40
							WriteLong texStream width
							WriteLong texStream height
							WriteShort texStream 1
							WriteShort texStream 8
							WriteLong texStream 0
							WriteLong texStream ( width * height)
							WriteLong texStream 0
							WriteLong texStream 0
							WriteLong texStream 0
							WriteLong texStream 0
							
							-- Color Palette
							for j = 1 to 256 do
							(
								local red = ReadByte abiStream
								local green = ReadByte abiStream
								local blue = ReadByte abiStream
								
								WriteByte texStream blue
								WriteByte texStream green
								WriteByte texStream red
								WriteByte texStream 0
							)
							
							fclose texStream
						)
						else
						(
							-- Skip Palette
							fseek abiStream 768 #seek_cur
						)
						fseek abiStream 40 #seek_cur -- skip Texture Name, Size, Offset
						
						-- Assign Bitmap
						newMaterial[i] = standard name:name
						newMaterial[i].diffuseMap = BitmapTexture()
						newMaterial[i].diffuseMap.name = name
						newMaterial[i].diffuseMap.filename = ( currentPath + directory + name )
						
						-- Show Map in Viewport
						showTextureMap newMaterial[i] true
						
					)
					
					-- BONES HEADER --
					local boneParent = #()
					local bone00 = #()
					local bone_xyz = #()
					local bone_name = #()
					local bone_n = #( )	
					
					for i = 1 to num_bone do
					(
						
						local  Parent_Index =ReadLong abiStream
						append boneParent Parent_Index 
						append bone00 ( ReadLong abiStream #unsigned )
						local x = ReadFloat abiStream
						local y = ReadFloat abiStream
						local z = ReadFloat abiStream
							
						append bone_xyz [ x,y,z ]
						append bone_name ( ReadCString abiStream 32 )
				
				
				
				   	append bone_n ( BoneSys.createBone [ x, y, z ] [ x+10, y, z ][ x, y, z ])
							bone_n[i].name = bone_name[i]										
				    
						  bone_n[i].width = 3
						   bone_n[i].height = 3

						  if Parent_Index != -1 then bone_n[i].parent = bone_n[Parent_Index+1] 
						
							 
				
				
					)
			
					
					
					-- MODELS HEADER --
					local model_num_vertex = #()
					local model_num_polygon = #()
					local model_offset_texure_m = #()
					local model_name = #()
					local model_offset_model_m = #()
					
					for i = 1 to num_model do
					(
						append model_num_vertex ( ReadShort abiStream #unsigned )
						append model_num_polygon ( ReadShort abiStream #unsigned )
						append model_offset_texure_m ( ReadShort abiStream #unsigned )
						append model_name ( ReadCString abiStream 32 )
						append model_offset_model_m ( ReadLong abiStream #unsigned )
					)
					
					local pos = ftell abiStream
					
					fseek abiStream offset_model #seek_set
					
					if convertCheckBox.checked then
					(
						writeLong outStream num_model
						writeLong outStream num_bone
					)
					
							
							
					-- DEBUG
					local newMesh
					local num_s
					local show_model = modelSpinner.value
					local show_animation = animationSpinner.value
					
			 	local StartVertex_arrss  = #()
				local EndVertex_arrss   = #()
		
					
					
					-- MODELS DATA --
					for i = 1 to num_model do
					(
						
						local verts = #()
						
      
						for j = 1 to model_num_vertex[i] do
						(
							local x = ReadFloat abiStream
							local y = ReadFloat abiStream
							local z = ReadFloat abiStream
							
							append verts [ x, y, z ]
						)
						
						local faces = #()
						local texFaces = #()
						local faceMaterialIDs = #()
						local borders_sum = 0
						local texcoords = #()
						
						local num_borders = #()
						local texture_id = #()
						local model_id = #()
						
						for j = 1 to model_num_polygon[i] do
						(
							
							local ID = #()
							local UV = #()
							
							append num_borders ( ReadByte abiStream #unsigned )
							append texture_id ( ReadShort abiStream #unsigned )
							
							for k = 1 to num_borders[j] do
							(
								append model_id ( ReadShort abiStream #unsigned )
								append ID ( model_id[model_id.count] + 1 )
								append UV ( borders_sum + k )
								local u = ( ( ReadShort abiStream #signed ) / 4096.0 )
								local v = ( ( ReadShort abiStream #signed ) / -4096.0 )
								append texcoords [ u, v, 1 ]
							)
							
							for k = 1 to num_borders[j]-2 do
							(
								append faces [ ID[k+2], ID[k+1], ID[1] ]
								append texFaces [ UV[k+2], UV[k+1], UV[1] ]
								append faceMaterialIDs ( texture_id[j] + 1 )
							)
							
							borders_sum += num_borders[j]
							
						)
						
						if convertCheckBox.checked then
						(
							borders_sum = 0
							
							writeLong outStream model_num_vertex[i]
							writeLong outStream model_num_polygon[i]
							writeCString outStream model_name[i] 32
							
							for j = 1 to model_num_vertex[i] do
							(
								writeFloat outStream verts[j][1]
								writeFloat outStream verts[j][2]
								writeFloat outStream verts[j][3]
							)
							
							for j = 1 to model_num_polygon[i] do
							(
								writeByte outStream num_borders[j]
								writeByte outStream texture_id[j]
								
								for k = 1 to num_borders[j] do
								(
									-- ID, u, v
									writeShort outStream model_id[borders_sum + k]
									writeShort outStream ( texcoords[borders_sum + k][1] * 4096.0 )
									writeShort outStream ( texcoords[borders_sum + k][2] * -4096.0 )
								)
								borders_sum += num_borders[j]
							)
						)
						
						-- BONES DATA --
						local StartVertex_arr  = #()
				       local EndVertex_arr   = #()  
						
						for j = 1 to num_bone do
						(
							
							local StartVertex  = ReadLong abiStream #unsigned
							local EndVertex  = ReadLong abiStream #unsigned
							
						append StartVertex_arr  StartVertex
						append EndVertex_arr    EndVertex
							
							
							if convertCheckBox.checked then
							(
								writeLong outStream StartVertex
								writeLong outStream EndVertex
							)
						)
						append StartVertex_arrss  StartVertex_arr
						append EndVertex_arrss    EndVertex_arr
						
						
						-- Construct the mesh object
						if i == show_model then
						(
						
						local newMesh = mesh name:model_name[i] pos:[0,0,0] \
										vertices:verts faces:faces \
										material:newMaterial
						
						-- Set face material IDs
						for j = 1 to faceMaterialIDs.count do
						(
							setFaceMatID newMesh j faceMaterialIDs[j]
						)
						
						-- Add texcoords
						if texcoords.count > 0 do
						(
							-- Add texcoords
							setNumTVerts newMesh texcoords.count
							
							for j = 1 to texcoords.count do
							(
								setTVert newMesh j texcoords[j]
							)
							
							-- Build texcoord faces
							buildTVFaces newMesh
							
							-- Set texcoord faces
							for j = 1 to texFaces.count do
							(
								setTVFace newMesh j texFaces[j]
							)
						)
						
						-- Update mesh
						update newMesh
					)
					)
					
					if convertCheckBox.checked then
					(
						-- write Bone Hierarchy
						for i = 1 to num_bone do
						(
							writeLong outStream boneParent[i]
							writeFloat outStream bone_xyz[i][1]
							writeFloat outStream bone_xyz[i][2]
							writeFloat outStream bone_xyz[i][3]
							writeCString outStream bone_name[i] 32
							writeLong outStream 0
						)
					)
					
	
	
	------------------	skin() ----			
					  max modify mode 
      		  vSkin = skin() 
			 
			      m=$alemet
				select m         
			   
			      addModifier m vSkin 
			      
			     
						for i=1 to bone_n.count do skinOps.addbone vSkin (getNodeByName bone_n[i].name ) -1
									   
					    skinOps.selectBone vskin 1
					--for i=1 to vertss.count do skinOps.ReplaceVertexWeights vskin i 1 1
						step = "Skinning complete"
					
					format "show_model %\n"  show_model
						for i=1 to bone_n.count do
							(
				
							
						
							
							if (StartVertex_arrss[show_model][i]) != (EndVertex_arrss[show_model][i]) then
							
							  (
							  	format "% %\n" (StartVertex_arrss[show_model][i]+1) (EndVertex_arrss[show_model][i])
							  for j = (StartVertex_arrss[show_model][i]+1) to (EndVertex_arrss[show_model][i]) do
								 (
									--format "% %\n" j i
 									 skinOps.ReplaceVertexWeights vSkin j i 1
						
								)
	    	  	)
	   		)
					
					
					fseek abiStream pos #seek_set
					
					-- ANIMATIONS DATA --
					local animation_name = #()
					local animation_num_s = #()
					local animation_unknown = #()
					
					for i = 1 to num_animation do
					(
						append animation_name ( ReadCString abiStream 32 )
						append animation_num_s ( ReadLong abiStream #unsigned )
						append animation_unknown ( ReadCString abiStream 16 )
					)
						num_s = animation_num_s[show_animation]
					pos = ftell abiStream
					
					fseek abiStream length_bone #seek_cur
					
					-- TIMEAXIS DATA --
					
					local offset_translateKeyframes = #()
					local offset_rotateKeyframes    = #()
					local num_translateKeyframes    = #()
					local num_rotateKeyframes       = #()
					local translateKeyframes        = #()
					local rotateKeyframes           = #()
					local translateTimestamps       = #()
					local rotateTimestamps          = #()
					
					local t_counter = 0
					local r_counter = 0
					
					
					
					
					
					
					for i = 1 to num_timeaxis do
					(
						local num_translateKeyframesi = ReadByte abiStream #unsigned
						local num_rotateKeyframesi = ReadByte abiStream #unsigned
						
						append offset_translateKeyframes t_counter
						append offset_rotateKeyframes r_counter
						
						append num_translateKeyframes num_translateKeyframesi
						append num_rotateKeyframes num_rotateKeyframesi 
						t_counter += num_translateKeyframes[i]
						r_counter += num_rotateKeyframes[i]
						
						local dx = #()
						local dy = #()
						local dz = #()
						local t_timestamp = #()
						
						for j = 1 to num_translateKeyframesi do
						(
							append dx ( readShort abiStream #signed )
							append dy ( readShort abiStream #signed )
							append dz ( readShort abiStream #signed )
						
												
	-------------------------------------------------- CHECK THIS --------------------------------------------------
							append translateKeyframes [ dx[j]/256.0, dy[j]/256.0, dz[j]/256.0 ]
	----------------------------------------------------------------------------------------------------------------
	
						
						
						)
						
						for j = 1 to num_translateKeyframesi do
						(
							append t_timestamp ( ReadByte abiStream #unsigned )
						 append translateTimestamps t_timestamp[j]
						
						)
						
						local rx = #()
						local ry = #()
						local rz = #()
						local rw = #()
						local r_timestamp = #()
						
						for j = 1 to num_rotateKeyframesi do
						(
							
				    	append rx (( ReadByte abiStream #signed )/128.0)
							append ry (( ReadByte abiStream #signed )/128.0)
							append rz (( ReadByte abiStream #signed )/128.0)
							
	-------------------------------------------------- CHECK THIS --------------------------------------------------
							append rw (sqrt(1- rx[j]*rx[j]-ry[j]*ry[j]-rz[j]*rz[j]))
							
							
							append rotateKeyframes ( Quat  rx[j] ry[j] rz[j] rw[j] )
	----------------------------------------------------------------------------------------------------------------
	
						)
						
						for j = 1 to num_rotateKeyframesi do
						(
							append r_timestamp ( ReadByte abiStream #unsigned )
					  	append rotateTimestamps r_timestamp[j]
						)
							
						if ( bit.and num_translateKeyframesi 1 ) != 0 then
						(
							local timeaxis64 = ReadByte abiStream #unsigned
						)
						
						if convertCheckBox.checked then
						(
							writeLong outStream num_translateKeyframes
							for j = 1 to num_translateKeyframes do
							(
								writeByte outStream t_timestamp[j]
								writeShort outStream dx[j]
								writeShort outStream dy[j]
								writeShort outStream dz[j]
							)
							
							writeLong outStream num_rotateKeyframes
							for j = 1 to num_rotateKeyframes do
							(
								writeByte outStream r_timestamp[j]
								
								-- TODO: Convert to correct values x,y,z,w
								local euler = eulerAngles rx[j] ry[j] rz[j]
								local q = eulerToQuat euler order:1
								
								writeShort outStream q.x
								writeShort outStream q.y
								writeShort outStream q.z
								writeShort outStream q.w
							)
						)
						
					)
					
					local pos2 = ftell abiStream
					
					fseek abiStream pos #seek_set
					
					-- RELATED BONES --
					for i = 1 to num_animation do
					(
						local num_related_bones = ReadLong abiStream #unsigned
						local bone_id = #()
						local transform_time_axis_idx = #()
						
						for j = 1 to num_related_bones do
						(
							append bone_id ( ReadLong abiStream #unsigned )
							append transform_time_axis_idx ( ReadLong abiStream #unsigned )
						)
						
						if convertCheckBox.checked then
						(
							writeCString outStream animation_name[i] 60
							writeLong outStream num_related_bones
							for j = 1 to num_related_bones do
							(
								writeLong outStream bone_id[j]
								writeLong outStream transform_time_axis_idx[j]
							)
						)
						if i == show_animation then
						
						(
							
							frameRate = 24
							animationRange = interval 0 num_s
							animate on
							(
								for j = 1 to num_related_bones do
								(
									if bone_n[bone_id[j]+1] != undefined then
									(
										
										-- TRANSLATION
										in coordsys parent op = copy bone_n[bone_id[j]+1].pos
										for k = 1 to num_translateKeyframes[transform_time_axis_idx[j]+1] do
										(
											at time translateTimestamps[offset_translateKeyframes[transform_time_axis_idx[j]+1] + k]
											(
												local tk = translateKeyframes[offset_translateKeyframes[transform_time_axis_idx[j]+1] + k]
												in coordsys parent bone_n[bone_id[j]+1].pos = op + tk
											)
										)
										
										-- ROTATION
										for k = 1 to num_rotateKeyframes[transform_time_axis_idx[j]+1] do
										(
											at time rotateTimestamps[offset_rotateKeyframes[transform_time_axis_idx[j]+1] + k]
											(
												local rk = rotateKeyframes[offset_rotateKeyframes[transform_time_axis_idx[j]+1] + k]
												setNodeWorldRotation bone_n[bone_id[j]+1] rk
											)
										)
										
									)
								)
							)
						)
						
					)
					
					fseek abiStream pos2 #seek_set
					
					-- TEXTURES DATA --
					for i = 1 to num_texture do
					(
						local name = texture_name[i]
						local width = texture_width[i]
						local height = texture_height[i]
						
						-- Extract Bitmap
						if extractTexturesCheckBox.checked then
						(
							local texName = currentPath + directory + name
							texStream = fopen texName "ab" -- append file
							
							local bitmap = #()
							
							for j = 1 to height do
							(
								local line = #()
								for k = 1 to width do
								(
									append line ( ReadByte abiStream )
								)
								append bitmap line
							)
							
							for j = 1 to height do
							(
								for k = 1 to width do
								(
									WriteByte texStream ( bitmap[height-j+1][k] )
								)
							)
							
							fclose texStream
							
						)
						else
						(
							-- skip Bitmap
							fseek abiStream ( width * height ) #seek_cur
						)
						
					)
					
				)
			
			
			)
			catch
			(
				messageBox "Error reading ABI file!"
				isValid = false
				fclose abiStream
				if convertCheckBox.checked then
					fclose outStream
				throw
			)
			
			
			-- Close the file stream
			fclose abiStream
			
			if convertCheckBox.checked then
				fclose outStream
			
			if isValid then
			(
				-- Redraw max viewports
				max views redraw
				
				print "Import completed."
			)
		)
	)
)