------------------------------------------------------------------
-- Commandos 3 ABI Import utility, v1.1							--
-- by GrafZeppelin (ferdinand.graf.zeppelin@gmail.com)			--
------------------------------------------------------------------

utility abiImport "C3 ABI - BONE ANIM" width:200 height:240
(
	
	--------------------
	-- User Interface --
	--------------------
	
	GroupBox grp1 "About" pos:[4,7] width:179 height:76
	label lab11 "Commandos 3" pos:[48,26] width:67 height:13
	label lab12 "ABI Importer v1.1" pos:[38,44] width:87 height:13
	HyperLink addy "by GrafZeppelin" pos:[44,62] width:75 height:15 address:"mailto:ferdinand.graf.zeppelin@gmail.com" color:(color 0 0 170) hovercolor:(color 170 0 0)
	
	GroupBox grp2 "Import" pos:[4,88] width:183 height:144
	button importButton "Import..." pos:[16,200] width:128 height:25
	
	label lbl3 "DEBUG:" pos:[16,112] width:104 height:16
	spinner modelSpinner "Model:       " pos:[20,133] width:137 height:16 range:[1,100,1] type:#integer scale:1
	spinner animationSpinner "Animation: " pos:[20,158] width:136 height:16 range:[1,100,2] type:#integer scale:1
	
	----------------------
	-- Helper Functions --
	----------------------
	
	-- Write a String which is not terminated by 0
	function WriteString file string =
	(
		for i = 1 to string.count do
		(
			WriteByte file ( bit.charAsInt string[i] )
		)
	)
	
	-- Write a String of a certain length
	function WriteCString file string length =
	(
		if string.count > length then
		(
			for i = 1 to length do
				WriteByte file ( bit.charAsInt string[i] )
			
			return 0
		)
		
		for i = 1 to string.count do
			WriteByte file ( bit.charAsInt string[i] )
		
		for i = string.count + 1 to length do
			WriteByte file 0
		
	)
	
	-- Read a C-String of a certain length
	function ReadCString stream length = 
	(
		local str = ""
		local b
		for i = 1 to length do
		(
			b = ReadByte stream
			if b == 0 then
			(
				fseek stream ( length - i ) #seek_cur
				return str
			)
			str += bit.intAsChar b
		)
		return str
	)
	
	fn setNodeWorldRotation theNode theRot = 
	(
		in coordsys (transmatrix theNode.transform.pos)
			theNode.rotation = theRot
	)
	
	--------------------------
	-- Main Import Function --
	--------------------------
	
	on importButton pressed do
	(
		
		-- Open up a dialog box
		abiFileName = getOpenFileName caption:"Import ABI" \
					types:"Commandos 3 ABI (*.abi)|*.abi|All Files (*.*)|*.*|"
		
		-- If user made a selection, begin importing
		if abiFileName != undefined then
		(
			local currentPath = getFilenamePath abiFileName
			
			-- Open the file
			abiStream = fopen abiFileName "rb"
			
			local string_LDMB = 1112360012
			local string_1011 = 825307185
			
				-- Check identifier
				if ( ReadLong abiStream ) != string_LDMB then
				(
					messageBox "Not a valid ABI file!"
				)
				-- Check version
				else if ( ReadLong abiStream ) != string_1011 then
				(
					messageBox "Not a Commandos 3 ABI file.\n\nTry to open it with the Commandos 2 importer!"
				)
				else
				(
					
					-- GENERAL INFO --
					local num_000 = ReadShort abiStream #unsigned
					local num_timeaxis = ReadShort abiStream #unsigned
					local num_animation = ReadShort abiStream #unsigned
					local num_texture = ReadShort abiStream #unsigned
					local num_model = ReadShort abiStream #unsigned
					local num_bone = ReadShort abiStream #unsigned
					
					local offset_texture = ReadLong abiStream #unsigned
					local offset_model = ReadLong abiStream #unsigned
					local offset_animation = ReadLong abiStream #unsigned
					local length_bone = ReadLong abiStream #unsigned
					local length_timeaxis = ReadLong abiStream #unsigned
					
					
					-- TEXTURES HEADER --
					local texture_width = #()
					local texture_height = #()
					local texture_name = #()
					local texture_size = #()
					local texture_offset = #()
					
					for i = 1 to num_texture do
					(
						
						append texture_width ( ReadShort abiStream #unsigned )
						append texture_height ( ReadShort abiStream #unsigned )
						
						fseek abiStream 768 #seek_cur -- skip Color Palette
						append texture_name ( ReadCString abiStream 32 )
						append texture_size ( ReadLong abiStream #unsigned )
						append texture_offset ( ReadLong abiStream #unsigned )
						fseek abiStream -808 #seek_cur -- go back to Color Palette
						
						local name = texture_name[i]
						
						-- Skip Palette
						fseek abiStream 768 #seek_cur
						
						fseek abiStream 40 #seek_cur -- skip Texture Name, Size, Offset
						
						
					)
					
					-- BONES HEADER --
					local bone_n = #()
					
					for i = 1 to num_bone do
					(
						local bone_parent = ReadLong abiStream #unsigned
						local bone00 = ReadLong abiStream #unsigned
						local x = ReadFloat abiStream
						local y = ReadFloat abiStream
						local z = ReadFloat abiStream
						local bone_name = ReadCString abiStream 32
						
						
						if bone_parent == -1 then
						(
							bone_n[i] = bone pos:[ x, y, z ]
						)
						else
						(
							in bone_n[bone_parent+1] bone_n[i] = bone pos:[ x, y, z ]
						)
						
						bone_n[i].name = bone_name
						append bone_n bone_n[i]
						
					)
					
					
					-- MODELS HEADER --
					local model_num_vertex      = #()
					local model_num_polygon     = #()
					local model_offset_texure_m = #()
					local model_name            = #()
					local model_offset_model_m  = #()
					
					for i = 1 to num_model do
					(
						append model_num_vertex ( ReadShort abiStream #unsigned )
						append model_num_polygon ( ReadShort abiStream #unsigned )
						append model_offset_texure_m ( ReadShort abiStream #unsigned )
						append model_name ( ReadCString abiStream 32 )
						append model_offset_model_m ( ReadLong abiStream #unsigned )
					)
					
					local pos = ftell abiStream
					
					fseek abiStream offset_model #seek_set
					
					
					-- DEBUG
					local newMesh
					local num_s
					local show_model = modelSpinner.value
					local show_animation = animationSpinner.value
					
					
					-- MODELS DATA --
					for i = 1 to num_model do
					(
						
						local verts = #()
						verts[model_num_vertex[i]] = 0
						
						for j = 1 to model_num_vertex[i] do
						(
							local x = ReadFloat abiStream
							local y = ReadFloat abiStream
							local z = ReadFloat abiStream
							
							append verts [ x, y, z ]
						)
						
						local faces = #()
						local texFaces = #()
						local faceMaterialIDs = #()
						local borders_sum = 0
						local texcoords = #()
						
						local num_borders = #()
						local texture_id = #()
						local model_id = #()
						
						for j = 1 to model_num_polygon[i] do
						(
							
							append num_borders ( ReadByte abiStream #unsigned )
							append texture_id ( ReadShort abiStream #unsigned )
							
							for k = 1 to num_borders[j] do
							(
								append model_id ( ReadShort abiStream #unsigned )
								local u = ( ( ReadShort abiStream #signed ) / 4096.0 )
								local v = ( ( ReadShort abiStream #signed ) / -4096.0 )
							)
							
							borders_sum += num_borders[j]
							
						)
						
						-- BONES DATA --
						for j = 1 to num_bone do
						(
							
							local StartVertex = ReadLong abiStream #unsigned
							local EndVertex = ReadLong abiStream #unsigned
							
						)
						
					)
					
					fseek abiStream pos #seek_set
					
					
					-- ANIMATIONS DATA --
					local animation_name = #()
					local animation_num_s = #()
					local animation_unknown = #()
					
					for i = 1 to num_animation do
					(
						append animation_name ( ReadCString abiStream 32 )
						append animation_num_s ( ReadLong abiStream #unsigned )
						append animation_unknown ( ReadCString abiStream 16 )
					)
					
					num_s = animation_num_s[show_animation]
					
					pos = ftell abiStream
					
					fseek abiStream length_bone #seek_cur
					
					
					-- TIMEAXIS DATA --
					local offset_translateKeyframes = #()
					local offset_rotateKeyframes    = #()
					local num_translateKeyframes    = #()
					local num_rotateKeyframes       = #()
					local translateKeyframes        = #()
					local rotateKeyframes           = #()
					local translateTimestamps       = #()
					local rotateTimestamps          = #()
					
					local t_counter = 0
					local r_counter = 0
					
					for i = 1 to num_timeaxis do
					(
						append offset_translateKeyframes t_counter
						append offset_rotateKeyframes r_counter
						
						append num_translateKeyframes ( ReadByte abiStream #unsigned )
						append num_rotateKeyframes ( ReadByte abiStream #unsigned )
						t_counter += num_translateKeyframes[i]
						r_counter += num_rotateKeyframes[i]
						
						local dx = #()
						local dy = #()
						local dz = #()
						local t_timestamp = #()
						
						for j = 1 to num_translateKeyframes[i] do
						(
							append dx ( readShort abiStream #signed )
							append dy ( readShort abiStream #signed )
							append dz ( readShort abiStream #signed )
							
	-------------------------------------------------- CHECK THIS --------------------------------------------------
							append translateKeyframes [ dx[j]/256.0, dy[j]/256.0, dz[j]/256.0 ]
	----------------------------------------------------------------------------------------------------------------
	
						)
						
						for j = 1 to num_translateKeyframes[i] do
						(
							append t_timestamp ( ReadByte abiStream #unsigned )
							append translateTimestamps t_timestamp[j]
						)
						
						local rx = #()
						local ry = #()
						local rz = #()
						local rw = #()
						local r_timestamp = #()
						
						for j = 1 to num_rotateKeyframes[i] do
						(
							
							append rx ( ReadByte abiStream #signed )
							append ry ( ReadByte abiStream #signed )
							append rz ( ReadByte abiStream #signed )
							
	-------------------------------------------------- CHECK THIS --------------------------------------------------
							append rotateKeyframes ( eulerToQuat ( eulerAngles rx[j] ry[j] rz[j] ) )
	----------------------------------------------------------------------------------------------------------------
	
						)
						
						for j = 1 to num_rotateKeyframes[i] do
						(
							append r_timestamp ( ReadByte abiStream #unsigned )
							append rotateTimestamps r_timestamp[j]
						)
							
						if ( bit.and num_translateKeyframes[i] 1 ) != 0 then
						(
							local timeaxis64 = ReadByte abiStream #unsigned
						)
						
					)
					
					
					local pos2 = ftell abiStream
					
					fseek abiStream pos #seek_set
					
					-- RELATED BONES --
					for i = 1 to num_animation do
					(
						local num_related_bones = ReadLong abiStream #unsigned
						local bone_id = #()
						local transform_time_axis_idx = #()
						
						for j = 1 to num_related_bones do
						(
							append bone_id ( ReadLong abiStream #unsigned )
							append transform_time_axis_idx ( ReadLong abiStream #unsigned )
						)
						
						
						if i == show_animation then
						(
							
							frameRate = 24
							animationRange = interval 0 num_s
							animate on
							(
								for j = 1 to num_related_bones do
								(
									if bone_n[bone_id[j]+1] != undefined then
									(
										
										-- TRANSLATION
										in coordsys parent op = copy bone_n[bone_id[j]+1].pos
										for k = 1 to num_translateKeyframes[transform_time_axis_idx[j]+1] do
										(
											at time translateTimestamps[offset_translateKeyframes[transform_time_axis_idx[j]+1] + k]
											(
												local tk = translateKeyframes[offset_translateKeyframes[transform_time_axis_idx[j]+1] + k]
												in coordsys parent bone_n[bone_id[j]+1].pos = op + tk
											)
										)
										
										-- ROTATION
										for k = 1 to num_rotateKeyframes[transform_time_axis_idx[j]+1] do
										(
											at time rotateTimestamps[offset_rotateKeyframes[transform_time_axis_idx[j]+1] + k]
											(
												local rk = rotateKeyframes[offset_rotateKeyframes[transform_time_axis_idx[j]+1] + k]
												setNodeWorldRotation bone_n[bone_id[j]+1] rk
											)
										)
										
									)
								)
							)
						)
						
					)
					
					fseek abiStream pos2 #seek_set
					
					-- TEXTURES DATA --
					for i = 1 to num_texture do
					(
						local name = texture_name[i]
						local width = texture_width[i]
						local height = texture_height[i]
						
						-- skip Bitmap
						fseek abiStream ( width * height ) #seek_cur
					)
					
				)
			
			fclose abiStream
			
			max views redraw
		)
	)
)